// 
//  Sprinkler controller with an ethernet interface
//
//  A stand alond sprinkler controller board, providing autonomous operation.  
//  Able to be programmed via an IP network.  
//
//  This version understands the Odd and Evens system used by some councils.....
//
//  This software uses the etherShield library and associated setup code.
//
//
// (C) 2010 Doug Jackson - Licensed under the TAPR Open Hardware License (www.tapr.org/NCL)
//
//  Extensive use has been made of the NuElectronics Ethershield example code
//
////////////////////////////////////////////////
//
//  V1.0 20100914 - DRJ - First Production Version (YAY! - that was fun!) WOOT!!!!
//
////////////////////////////////////////////////
//
//
// Define this to include additional debug messages
//#define DEBUG_PRINT


#include "etherShield.h"
#include <stdio.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <EEPROM.h>
#include <DS1302.h>

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {
  0x54,0x55,0x58,0x10,0x00,0x25}; 
static uint8_t myip[4] = {
  192,168,1,2};
static char  baseurl[]="http://192.168.1.2/";
static uint16_t mywwwport =80; // listen port for tcp/www (max range 1-254)

prog_char header[] PROGMEM ="HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n";
prog_char title[] PROGMEM ="<center><p><h1>Sprinkler Controller</h1></p> ";
prog_char command_line[] PROGMEM ="<h3><a href=\"/?cmd=1\">Program</a> &nbsp;<a href=\"/?cmd=2\">Time</a> &nbsp;<a href=\"/?cmd=4\">Manual</a>&nbsp;</h3>";
prog_char djcopyright[] PROGMEM = "<p>ACT Government Compliant Version - V1.0 - (C)2010 Doug Jackson";

// create a string table pointing to our PROGMEM strings (Anything to save RAM!!!)
PROGMEM const char *string_table[]=
{ 
  header,
  command_line,
  djcopyright,
  title
};

// an array of numbers that can be used to identify the number of days in the previous month
static uint8_t days_in_previous_month[12]={
  31,31,28,31,30,31,30,31,31,30,31,30};

Time time;


// Ethernet buffers
// These are very problematic - Have to make sure that each web page returned is
// smaller than this buffer size.  Increasing the buffer will impact on stack space
// and everything will stop!! - You have been warned!
#define BUFFER_SIZE 870    // should be 910
static uint8_t buf[BUFFER_SIZE+1];


// A structure to hold our programs
// one entry per solenoid output
typedef struct program_internal
{
  uint8_t water_time_remaining;  // amount of time left to water this zone
  uint8_t water_time;            // the programmed amount of time to water
  uint8_t water_days;            // day cycle for this zone - every n days
  uint8_t last_watered;          // the day number we were last watered
  uint8_t water_status;          // 0=do not water    1=currently watering 
  // 2=finished watering today  3 = scheduled for watering
} 
programs;

#define DO_NOT_WATER      0
#define WATERING          1
#define FINISHED_WATERING 2
#define SCHEDULED         3

#define NUM_PROGS 5
programs progs[NUM_PROGS];


// define the layout of EEPROM storage
#define EEPROM_SIGNATURE        0
#define EEPROM_RUNSTATE         1
#define EEPROM_STARTHOUR        2
#define EEPROM_STARTMINUTE      3
#define EEPROM_WATERINGDAY      4
#define EEPROM_STRUCTURE_OFFSET 5




uint8_t watering_day;              // 1 = odd, 0 = even
uint8_t start_hour, start_minute;  // storage for the start times for our watering program
static uint8_t current_state;      // The storage for our current state
uint16_t r;                        // a return value register used by various processes - defined as a global so each routine
                                   // wasn't creating one!


// a temporary string buffer
#define LINE_BUFFER_SIZE 180
char line_buffer[LINE_BUFFER_SIZE];


EtherShield es=EtherShield();


// Hardware definitions to match the new PCB.
// The actual solenoud outputs are connected to the following pins;
#define OUTPUT1 9
#define OUTPUT2 5
#define OUTPUT3 6
#define OUTPUT4 7
#define OUTPUT5 8
#define OUTPUT6 4
uint8_t OUTPUTS[6]={
  OUTPUT1, OUTPUT2, OUTPUT3, OUTPUT4, OUTPUT5, OUTPUT6};

// The RTC interface to the DS1302
#define RTC_R        19
#define RTC_IO       18
#define RTC_CLK      17

// the RainLED is on Digital 3 - pin 5
#define rainLED 3

// the Rain button is on Analog 0 - Digital 14
#define rainButton 14

DS1302 rtc(RTC_R, RTC_IO, RTC_CLK);



static uint8_t output_status;  // A variable to store which outputs are on at the moment
byte run_status;            // holds the status of our timer - 1 = running 0 = on hold for manual operation
byte rain_status;           // is it raining?  1=yes, 0=no 
uint8_t currently_watering;


void do_zone_selection_ACT_GOVT()
{
  uint8_t date = rtc.date();
  //  (1)  If (it is our defined day (odd/even) to operate) OR (day != 31), and the current time 
  //     matches our defined start time
  if( ((date%2 == watering_day) | date==31) & (start_hour == rtc.hour()) & (start_minute==rtc.minutes()) )
  {

    //debug output
#ifdef DEBUG_PRINT
    Serial.println("before:");
    debug_print_program_struct();
#endif

    //  (2)  for every zone in the zone list:
    for (int zone=0; zone<=4; zone++)
    {   
      // (3)  Identify a zone to run by
      //  (a)  If TodaysDayNumber < the LastRun (ie we are at the start of a new month)
      if (rtc.date()< progs[zone].last_watered)
      {
        //  calculate a WaitNumber = DaysInLastMonth-LastRun
        int waitnumber= (int)days_in_previous_month[rtc.month()-1] - (int)progs[zone].last_watered;
     
        //   if TodaysDayNumber + WaitNumber >= RunFrequency, AND the zone has some run tie in the program
        if (((rtc.date() + waitnumber) >= progs[zone].water_days) & (progs[zone].water_time>0) )
        {
          //   run the specified station by updating the LastRun date for the station
          progs[zone].last_watered=rtc.date();
          //   updating the water time remaining
          progs[zone].water_time_remaining=progs[zone].water_time;
          //   and setting the zones SCHEDULED flag
          progs[zone].water_status=SCHEDULED;
        }
      }

      //  (b)  If TodaysDayNumber > LastRun 
      if (rtc.date() > (int)progs[zone].last_watered)
      {
        // if TodaysDayNumber-RunFreq >= LastRun  AND the zone has some run time in the program 
        if (((rtc.date()-progs[zone].water_days) >= progs[zone].last_watered) & (progs[zone].water_time>0))
        {
          //   run the specified station by updating the LastRun date for the station
          progs[zone].last_watered=rtc.date();
          //   updating the water time remaining
          progs[zone].water_time_remaining=progs[zone].water_time;
          //   and setting the zones SCHEDULED flag
          progs[zone].water_status=SCHEDULED;
        }
      } 

    } 
    //debug output
#ifdef DEBUG_PRINT
    Serial.println("after:");
    debug_print_program_struct();
#endif
  }
}


void control_watering_solenoids()
{
  //Serial.println("Debug-1");
  //   if a zone is in WATERING state, decrement the water_time_remaining counter
  //     if the counter hits 00 then turn off the solenoid, and set the water_time_remainig 
  //        to what was stored in eeprom, set the FINISHED_WATERING flag

  currently_watering=0; // assume that we are not watering
  for (int zone=0; zone<=4; zone++)
  {   
    //Serial.print("z"); Serial.print(zone); Serial.print(":"); Serial.println((int)progs[zone].water_time_remaining);
    if (progs[zone].water_status==WATERING) 
    {
      currently_watering=1;  // we are currently watering
      if ( (progs[zone].water_time_remaining--) == 0)
      {
        output_status &=  ~(1<<zone);  // turn off the zone
        currently_watering=0;          // declare the fact that we are not watering any more
        output_status &= ~(1<<5);     // turn off the pump.
        progs[zone].water_status=FINISHED_WATERING;
        progs[zone].water_time_remaining=progs[zone].water_time;
      }
      else 
      {
        output_status |=  (1<<zone); // ensure that the zone is on!
        output_status |=  (1<<5);    // turn on the pump.
      }
    }
  }

  //   if no zones are in WATERING state, 
  if (currently_watering==0)  // we are not watering at the moment
    //  search for a zone that is in SCHEDULED state, then turn it on, 
    //    and set it to WATERING stsre
    for (int zone=0; zone<=4; zone++)
    {
      if (currently_watering==0)
        if (progs[zone].water_status==SCHEDULED) 
        {
          progs[zone].water_status=WATERING;
          currently_watering=1;  // set the flag to say we are watering - that way we will only turn on one zone
          progs[zone].water_time_remaining--;  // decrement the current water time remaining counter
          // to account for this minutes worth of watering
          output_status |=  (1<<zone);         // ensure that the zone is on!
          output_status |=  (1<<5);    // turn on the pump.
        } 
    }
}






void setup()
{
  uint8_t i;

  Serial.begin(9600);
  //Serial.println("Ethernet Sprinkler Controller v1.0");

  //if(BUFFER_SIZE !=910) Serial.println("WARNING - ETHERNET BUFFER IS NOT CORRECT....");

  // setup the outputs
  Serial.println("Reset outputs");
  for (i=0; i<6; i++)  pinMode(OUTPUTS[i], OUTPUT);  // start with the sprinklers turned off
  output_status=0;

  // set the Rain LED as an output
  pinMode(rainLED, OUTPUT);

  // set the Rain Button as an input - with the Internal Pullup high
  pinMode(rainButton, INPUT);
  digitalWrite(rainButton, HIGH);


  Serial.println("Init Ethernet");
  //initialize enc28j60
  es.ES_enc28j60Init(mymac);
  es.ES_enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
  delay(10);

  // Magjack leds configuration, see enc28j60 datasheet, page 11 
  // LEDA=greed LEDB=yellow
  // 0x880 is PHLCON LEDB=on, LEDA=on
  // enc28j60PhyWrite(PHLCON,0b0000 1000 1000 00 00);
  es.ES_enc28j60PhyWrite(PHLCON,0x880);
  delay(500);
  //
  // 0x990 is PHLCON LEDB=off, LEDA=off
  // enc28j60PhyWrite(PHLCON,0b0000 1001 1001 00 00);
  es.ES_enc28j60PhyWrite(PHLCON,0x990);
  delay(500);
  //
  // 0x880 is PHLCON LEDB=on, LEDA=on
  // enc28j60PhyWrite(PHLCON,0b0000 1000 1000 00 00);
  es.ES_enc28j60PhyWrite(PHLCON,0x880);
  delay(500);
  //
  // 0x990 is PHLCON LEDB=off, LEDA=off
  // enc28j60PhyWrite(PHLCON,0b0000 1001 1001 00 00);
  es.ES_enc28j60PhyWrite(PHLCON,0x990);
  delay(500);
  //
  // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
  // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
  es.ES_enc28j60PhyWrite(PHLCON,0x476);
  delay(100);

  //init the ethernet/ip layer:
  es.ES_init_ip_arp_udp_tcp(mymac,myip,80);




  if(EEPROM.read(0) != 0xa5)
  {
    Serial.println("EEPROM unitialized");

    EEPROM.write(EEPROM_SIGNATURE, 0xa5);    // Signature
    EEPROM.write(EEPROM_RUNSTATE, 0x01);     // 1 = Running on normal timer..   0 = halted for manual use
    EEPROM.write(EEPROM_STARTHOUR, 10);      // start Hour
    EEPROM.write(EEPROM_STARTMINUTE, 20);    // start Minute
    EEPROM.write(EEPROM_WATERINGDAY, 1);     // Watering day - 1=odd, 0=even


#define NUM_DEFAULT_PROGS 5

    progs[0].water_time = 2;
    progs[0].water_time_remaining = 2;    
    progs[0].water_days = 1;
    progs[0].last_watered=1;
    progs[0].water_status=0;
    progs[1].water_time = 2;
    progs[1].water_time_remaining = 2;    
    progs[1].water_days = 10;
    progs[1].last_watered=29;
    progs[1].water_status = 0;
    progs[2].water_time = 0;
    progs[2].water_time_remaining = 0;    
    progs[2].water_days = 0;
    progs[2].last_watered=0;
    progs[2].water_status = 0;
    progs[3].water_time = 0;
    progs[3].water_time_remaining = 0;    
    progs[3].water_days = 0;
    progs[3].last_watered=0;
    progs[3].water_status = 0;
    progs[4].water_time = 0;
    progs[4].water_time_remaining = 0;    
    progs[4].water_days = 0;
    progs[4].last_watered=0;
    progs[4].water_status = 0;


    for(i = 0; i < sizeof(programs) * NUM_DEFAULT_PROGS; i++)
    {
      EEPROM.write(EEPROM_STRUCTURE_OFFSET + i, ((uint8_t *) progs)[i]);
    }
  }
  else
  {
    Serial.println("Read PROM");
    run_status = EEPROM.read(1);
    //if (run_status==0) Serial.println("Stop"); else Serial.println("Run");

    for(i = 0; i < sizeof(programs) * NUM_PROGS; i++)
      ((uint8_t *) progs)[i] = EEPROM.read(EEPROM_STRUCTURE_OFFSET + i);
  }


  // grab the current time from the RTC to see if it is invalid...
  // time = rtc.time();

  if(  (rtc.read_register(0x20) != 0xa5) | 
    (rtc.date()<0) | (rtc.date()>31) | 
    (rtc.year()<2010) | (rtc.year()>2200) )
  {
    //Serial.println("RTC BROKE");
    rtc.write_protect(false);
    rtc.halt(false);
    Time t(2010, 9, 01, 22, 10, 0, 7);
    rtc.time(t);
    rtc.write_register(0x20, 0xa5);
  }

  if(rtc.read_register(0x20) == 0xa5)
  {
    //Serial.println("RTC present");
    rtc.write_protect(false);
    rtc.halt(false);
    rtc.write_register(0x09, 169);
    //    time = rtc.time();

  } //else Serial.println("RTC absent");


  // now we are running - flash the Rain LED 3 times to say we are happy!
  for (int i=0; i<3; i++) {
    digitalWrite(rainLED, HIGH);
    delay(500);
    digitalWrite(rainLED, LOW);
    delay(500);
  } 

  //progs[2].water_status=WATERING; progs[2].water_time_remaining=2; // debug
  //progs[3].water_status=SCHEDULED; progs[3].water_time_remaining=2; // debug

  // get the time and day we are suposed to start watering from EEPROM
  start_hour=EEPROM.read(EEPROM_STARTHOUR);
  start_minute=EEPROM.read(EEPROM_STARTMINUTE);
  watering_day = EEPROM.read(EEPROM_WATERINGDAY);

  //debug output
#ifdef DEBUG_PRINT
  Serial.print("Start_hr="); 
  Serial.println((int)start_hour);
  Serial.print("Start_min="); 
  Serial.println((int)start_minute);
  Serial.print("water_day="); 
  Serial.println((int)watering_day);
  rtc.minutes(start_minute-1);
  rtc.hour(start_hour);
  Serial.print("current_day="); 
  Serial.println((int)rtc.date());
  rtc.seconds(50);
  rtc.date(3);
#endif
}


void loop(){
  uint16_t plen, dat_p;
  int8_t cmd, i;

  static uint8_t armed = 0;
  currently_watering = 0;


  // debug functions to print a second counter
  //debug output
#ifdef DEBUG_PRINT
  int last_sec =0;
  last_sec=rtc.seconds();
  delay(200);
  if (rtc.seconds()!=last_sec) Serial.println((int)last_sec);
#endif

  // get the time and day we are suposed to start watering from EEPROM
  start_hour=EEPROM.read(EEPROM_STARTHOUR);
  start_minute=EEPROM.read(EEPROM_STARTMINUTE);
  watering_day = EEPROM.read(EEPROM_WATERINGDAY);


  // do the zone processing here, before we try to do the webpage processing...
  // only do the timer based functions if the run status is on! 
  if (run_status!=0) {

    // determine what needs to be watered 
    //  (0)  if it is raining - dont proceed.
    if (rain_status==0)
    {  
      // test function to verify the hardware works
      if (digitalRead(rainButton)==0) digitalWrite(rainLED,HIGH); 
      else digitalWrite(rainLED,LOW);

      // test the watering status every minute as we transition from 1 - 2
      if (rtc.seconds() == 1)  armed = 1; 
      if ( (rtc.seconds() == 2) & (armed == 1)) // we are at a 1 - 2 second boundary - 
      {
        armed=0;

        // identify zones to start
        do_zone_selection_ACT_GOVT();
        
        // actually start them
        control_watering_solenoids();

      }
      // update the output status if we are running.
      for (i=0; i<6; i++) 
      {
        if ((lowByte(output_status) & 1<<i)==0) digitalWrite(OUTPUTS[i], LOW); 
        else digitalWrite(OUTPUTS[i], HIGH);
      }
    }   

  }









  // get a packet if there is one available
  plen = es.ES_enc28j60PacketReceive(BUFFER_SIZE, buf);

  //if (plen!=0) Serial.println(plen);

  //plen will be unequal to zero if there is a valid packet (without crc error) 
  if(plen!=0){           
    // arp is broadcast if unknown but a host may also verify the mac address by sending it to a unicast address.
    if(es.ES_eth_type_is_arp_and_my_ip(buf,plen)){
      //Serial.println("ARP");
      es.ES_make_arp_answer_from_request(buf);
      return;
    }

    // check if ip packets are for us:
    if(es.ES_eth_type_is_ip_and_my_ip(buf,plen)==0){
      return;
    }

    if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
      es.ES_make_echo_reply_from_request(buf,plen);
      return;
    }

    // tcp port www start, compare only the lower byte
    if (buf[IP_PROTO_P]==IP_PROTO_TCP_V&&buf[TCP_DST_PORT_H_P]==0&&buf[TCP_DST_PORT_L_P]==mywwwport){
      if (buf[TCP_FLAGS_P] & TCP_FLAGS_SYN_V){
        es.ES_make_tcp_synack_from_syn(buf); // make_tcp_synack_from_syn does already send the syn,ack
        return;     
      }
      if (buf[TCP_FLAGS_P] & TCP_FLAGS_ACK_V){
        es.ES_init_len_info(buf); // init some data structures
        dat_p=es.ES_get_tcp_data_pointer();
        if (dat_p==0){ // we can possibly have no data, just ack:
          if (buf[TCP_FLAGS_P] & TCP_FLAGS_FIN_V){
            es.ES_make_tcp_ack_from_any(buf);
          }
          return;
        }


        if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
          plen=print_manual_page(buf);
          goto SENDTCP;
        }

        cmd=get_key_val((char *)&(buf[dat_p+5]),"cmd");
        if (cmd==1){
          //Serial.println("CMD1");
          plen=print_program_page(buf);
        } 

        if (cmd==2){
          uint8_t hour=get_key_val((char *)&(buf[dat_p+5]),"hr"); 
          uint8_t minute=get_key_val((char *)&(buf[dat_p+5]),"mn"); 
          uint8_t day=get_key_val((char *)&(buf[dat_p+5]),"dy"); 
          uint8_t month=get_key_val((char *)&(buf[dat_p+5]),"mo"); 
          uint16_t year= get_key_val((char *)&(buf[dat_p+5]),"yr"); 
          uint8_t shour= get_key_val((char *)&(buf[dat_p+5]),"sh"); 
          uint8_t sminute= get_key_val((char *)&(buf[dat_p+5]),"sm"); 
          uint8_t od= get_key_val((char *)&(buf[dat_p+5]),"od"); 
          if ( (hour>=0) & (hour<=23) & (minute>=0) & (minute<=59))  
          {
            rtc.hour(hour); 
            rtc.minutes(minute);
          }
          if ( (day>=0) & (day<=31) & (month>=0) & (month<=12) & (year>=2010) & (year<=2200))  
          {
            rtc.date(day); 
            rtc.month(month);
            rtc.year(year);
          } 
          if ( (shour>=0) & (shour<=23) & (sminute>=0) & (sminute<=59))  
          {
            start_hour=shour; 
            start_minute=sminute;
            watering_day=od;
            EEPROM.write(EEPROM_STARTHOUR, shour);      // start Hour
            EEPROM.write(EEPROM_STARTMINUTE, sminute);  // start Minute
            EEPROM.write(EEPROM_WATERINGDAY, od);       // Watering day - 1=odd, 0=even

          }

          plen=print_time_page(buf);
        }    




        if (cmd==4){
          uint8_t tmr=(uint8_t)get_key_val((char *)&(buf[dat_p+5]),"tmr"); // if there is a run status command - set that accordingly
          if (tmr==1) run_status=0;
          if (tmr==2) run_status=1;
          uint8_t rrly=(uint8_t)get_key_val((char *)&(buf[dat_p+5]),"zn"); // if there is a zone set command - toggle the appropriate output 
          if (rrly>=1)  output_status^=1<<--rrly;  // toggle the state of the output bit
          EEPROM.write(1, run_status);   

          for (i=0; i<6; i++) {
            if ((lowByte(output_status) & 1<<i)==0) digitalWrite(OUTPUTS[i], LOW); 
            else digitalWrite(OUTPUTS[i], HIGH);
          }
          plen=print_manual_page(buf);
        }
        
        if (cmd==5){
          uint8_t wtime=(uint8_t)get_key_val((char *)&(buf[dat_p+5]),"tm"); // if there is a time command, grab it
          uint16_t wday=get_key_val((char *)&(buf[dat_p+5]),"dy"); // if there is a day value, grab it 
          uint8_t zone=(uint8_t)get_key_val((char *)&(buf[dat_p+5]),"zn");  // if there is a zone value, grab it 

          // test the parameters that were passed in - if they are legal then change the program
          if ( ((wtime!=-1) & (wday!=-1)) & ((zone>=1) & (zone<=5)) ) {

            //sprintf(line_buffer, "Doing program table change - zone=%u, time=%d, day=%d, zone, wtime, wday);
            //Serial.println(line_buffer);

            progs[zone-1].water_time = wtime; 
            progs[zone-1].water_days=wday;
            progs[zone-1].water_status=0;
            for(i = 0; i < sizeof(programs) * NUM_PROGS; i++)
            {
              EEPROM.write(EEPROM_STRUCTURE_OFFSET + i, ((uint8_t *) progs)[i]);
            }

          }

          if (zone<1 | zone>5) zone=1;

          plen=edit_program_page(buf,zone);
        }        

SENDTCP:  
        es.ES_make_tcp_ack_from_any(buf); // send ack for http get
        es.ES_make_tcp_ack_with_data(buf,plen); // send data       
      }
    }
  }       



}



// The returned value is stored in the global var strbuf
uint8_t find_key_val(char *str,char *key)
{
  uint8_t found=0;
  uint8_t i=0;
  char *kp;
  kp=key;
  while(*str &&  *str!=' ' && found==0){
    if (*str == *kp){
      kp++;
      if (*kp == '\0'){
        str++;
        kp=key;
        if (*str == '='){
          found=1;
        }
      }
    }
    else{
      kp=key;
    }
    str++;
  }
  if (found==1){
    // copy the value to a buffer and terminate it with '\0'
    while(*str &&  *str!=' ' && *str!='&' && i<LINE_BUFFER_SIZE){
      line_buffer[i]=*str;
      i++;
      str++;
    }
    line_buffer[i]='\0';
  }
  return(found);
}

int16_t get_key_val(char *str, char* cmd)
{
  r=-1;
  if (find_key_val(str,cmd)){
    if (*line_buffer < 0x3a && *line_buffer > 0x2f){
      // is a ASCII number, return it
      //r=(*line_buffer-0x30);
      r = atoi(line_buffer);
    }
  }
  return r;
}


void debug_print_program_struct()
{
  for (int count=0; count<=4; count++) {

    sprintf(line_buffer, "Zone:%u WT:%u WTR:%u WD:%u LW:%u WS:%u", 
    count, progs[count].water_time,  progs[count].water_time_remaining, 
    progs[count].water_days, progs[count].last_watered, progs[count].water_status);
    Serial.println(line_buffer);
  }
}


uint16_t print_program_page(uint8_t *buf)
{

  uint8_t tempcount=0, count=0, rlycnt=0;
  uint8_t start=1, end=5;

  uint16_t plen;

  //plen=es.ES_fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n"));
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[0]))); // get header string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[3]))); // get title string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<center><p><h1>Sprinkler Controller"));
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[1]))); // get command string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h2>Current Program "));
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h1>"));
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<table width=\"300\" border=1><tr><th>Zone<p>#<th>Water Time<p>(Minutes)<th>Water every x days<p>2,4,6,8,10</tr>"));

  for (count=start; count<=end; count++) {

    sprintf(line_buffer, "<tr align=\"center\"><td>%u<td>%u<td>%u</tr>", 
    count, progs[count-1].water_time,  progs[count-1].water_days);
    // Serial.println(line_buffer);


    tempcount=0;
    while (line_buffer[tempcount]) {
      buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
      plen++;
    }

  }

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</table>") );

  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<p><table width=\"300\" border=1><tr><th>Start at<p>"));

  sprintf(line_buffer, "Start at: %d:%d",  start_hour, start_minute);
  // Serial.println(line_buffer);


  tempcount=0;
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h4><form method=GET action=/>Zone:<input type=text name=zn>") );
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=cmd value=5>"));     
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<input type=\"submit\" value=\"Modify\" /></form>") );

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</center><hr>"));
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[2]))); // get copyright string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  Serial.print("plen=");
  Serial.println(plen);

  return(plen);
}



//  /?cmd=4    Manual settings page - Make manual setting changes to solenoids and allow timer to be turned on or off
//  /?cmd=4?zn=1?tmr=0 - format for returning manual settings toggles solenoid1 state and turns off timer
uint16_t print_manual_page(uint8_t *buf)
{
  uint8_t tempcount=0, count=0;

  uint16_t plen;

  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[0]))); // get header string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[3]))); // get title string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<center><p><h1>Sprinkler Controller"));
  
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[1]))); // get command string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h2>Manual Control"));

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<table border=\"1\" width=\"200\"><tr><th>Zone<th>Status<th>Toggle"));
  for (count=0; count<=5; count++) {
    if ((lowByte(output_status) & (1<<count))==0)
    {
      if (count==5) 
        sprintf(line_buffer, "<tr><td>Pump<td>OFF<td><a href=/?cmd=4?zn=%d>Toggle</a>", count+1);
      else
        sprintf(line_buffer, "<tr><td>%d<td>OFF<td><a href=/?cmd=4?zn=%d>Toggle</a>", count+1, count+1);
    }
    else
    {
      if( count==5)
        sprintf(line_buffer, "<tr><td>Pump<td>ON<td><a href=/?cmd=4?zn=%d>Toggle</a>", count+1);
      else
        sprintf(line_buffer, "<tr><td>%d<td>ON<td><a href=/?cmd=4?zn=%d>Toggle</a>", count+1, count+1);

    }

    tempcount=0;
    while (line_buffer[tempcount]) {
      buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
      plen++;
    }

  }

  if (run_status==0) plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<tr><th>Timer<td>OFF<td><a href=/?cmd=4?tmr=2>ON</a>") );
  else plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<tr><th>Timer<td><a href=/?cmd=4?tmr=1>OFF</a><td>ON") );

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</table><br>&nbsp;<p>  ") );
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</center><hr>"));
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[2]))); // get copyright string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  Serial.print("plen=");
  Serial.println(plen);


  return(plen);
}


//  /?cmd=5?zn=1   Edit Program page  - Allows changes to be made to the state settings
//  /?cmd=5?zn=1?tm=2?dy=2 - format for returning state changes
uint16_t edit_program_page(uint8_t *buf, uint8_t zone)
{

  uint8_t tempcount=0, count=0, rlycnt=0;
  uint8_t start=0, end=0;

  uint16_t plen;

  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[0]))); // get header string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }

  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[3]))); // get title string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<center><p><h1>Sprinkler Controller"));
  
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[1]))); // get command string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h2> Modify Program </h2> "));
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h1>Current Settings:<br />"));
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<table border=1><tr><th>Zone<p>#<th>Water Time<p>(Minutes)<th>Water every x days (2,4,6,8,10)</tr>"));


  sprintf(line_buffer, "<tr align=center><td>%u<td>%u<td>%u</tr>", 
  zone, progs[zone-1].water_time, progs[zone-1].water_days);
  // Serial.println(line_buffer);        

  tempcount=0;
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</table>") );

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h4><form method=GET action=/>Water Time:<input type=text name=tm>Water Days:<input type=text name=dy><br/>") );

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<br />") );  
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=cmd value=5>"));
  sprintf(line_buffer, "<input type=hidden name=zn value=%d>", zone );
  // Serial.println(line_buffer);        

  tempcount=0;
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<input type=\"submit\" value=\"Modify\" /></form>") );


  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</center><hr>"));
  //tempcount=0;
  //strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[2]))); // get copyright string 
  //while (line_buffer[tempcount]) {
  //           buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
  //           plen++;
  //           }
  Serial.print("plen=");
  Serial.println(plen);

  return(plen);
} 


//  /?cmd=2   print time page 
//  /?cmd=2?hr=1?mn=2?next=2?dy=02?mo=02?yr=2010 - format for returning state changes
uint16_t print_time_page(uint8_t *buf)
{

  uint8_t tempcount=0, count=0, rlycnt=0;
  uint8_t start=0, end=0;

  uint16_t plen;

  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[0]))); // get header string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }

  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[3]))); // get title string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<center><p><h1>Sprinkler Controller"));

  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[1]))); // get command string 
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h2> Modify Program </h2> "));
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h1>Current Settings:<br />"));
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<table border=1><tr><th>Date<th>Time</tr>"));


  sprintf(line_buffer, "<tr align=center><td>%02d/%02d/%04d<td>%02d:%02d:%02d</tr>", 
  rtc.date(), rtc.month(), rtc.year(), rtc.hour(), rtc.minutes(), rtc.seconds());


  tempcount=0;
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</table>") );

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<table border=1><tr><th>Start Hour<th>Start Min<th>Odd/Even</tr>"));
    sprintf(line_buffer, "<tr align=center><td>%02d<td>%02d<td>%1d</tr>", 
         start_hour, start_minute, watering_day?1:0);

        
  tempcount=0;
  while (line_buffer[tempcount]) {
    buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
    plen++;
  }

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</table>") );
 
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h4><form method=GET action=/>Hour:<input type=text name=hr>Min:<input type=text name=mn><br/>") );
  //plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("Day:<input type=text name=dy>Month:<input type=text name=mo>Year:<input type=text name=yr>") );
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<h4><form method=GET action=/>Hour:<input name=hr>Min:<input name=mn><br/>") );
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("Day:<input name=dy>Month:<input name=mo>Year:<input name=yr><br/>") );
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("Start Hour:<input name=sh>Start Min:<input name=sm> Odd/Even:<input name=od>") );
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<br />") );  
  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=cmd value=2>"));

  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("<input type=\"submit\" value=\"Modify\" /></form>") );


  plen=es.ES_fill_tcp_data_p(buf,plen,PSTR("</center><hr>"));
  tempcount=0;
  strcpy_P(line_buffer, (char*)pgm_read_word(&(string_table[2]))); // get copyright string 
  while (line_buffer[tempcount]) {
             buf[TCP_CHECKSUM_L_P+3+plen]=line_buffer[tempcount++];
             plen++;
             }
  Serial.print("plen=");
  Serial.println(plen);

  return(plen);
} 


